
/**********************************************************************************************************************************/
/***************************************************  Useful routines for CBEM  ***************************************************/
/**********************************************************************************************************************************/

/* Developer: Sylvain Nintcheu Fata */
/* Version:   V.1.3                 */
/* Released:  Jun 14, 2012          */


#include <math.h>


double gammai(double p1, double p2, double qet, double cr1, double cr2, double zn, double zd)
{

/* ===purpose=== */

/**********************************************************************************************************************************/
/***************************************************  Determine gamma_i  **********************************************************/
/**********************************************************************************************************************************/

/* Developer: Sylvain Nintcheu Fata, Oak Ridge National Laboratory */
/* Version:   V.1.2                                                */
/* Released:  Apr 14, 2011                                         */


   extern double pix2; /* pix2 = 2*pi, pi = 3.1415926535897932 */



	     if ( (p1<0.0) && (p2>0.0) ) /* 1-----x----------2 <==> if Projection of collocation point falls in side k */
	     {
		         if ( qet<0.0 )
			 {
	                      if ( (cr1>0.0) && (cr2<0.0) && (zn>0.0) && (zd<0.0) )
		                return atan2(zn,zd) - pix2;
		              else if ( (cr1<0.0) && (cr2>0.0) && (zn>0.0) && (zd<0.0) )
		                return atan2(zn,zd) - pix2;
		              else if ( (cr1<0.0) && (cr2<0.0) && (zn>0.0) )
		                return atan2(zn,zd) - pix2;
		              else
		                return atan2(zn,zd);
		         }
			 else if ( qet>0.0 )
			 {
	                      if ( (cr1>0.0) && (cr2<0.0) && (zn<0.0) && (zd<0.0) )
		                 return atan2(zn,zd) + pix2;
		              else if ( (cr1<0.0) && (cr2>0.0) && (zn<0.0) && (zd<0.0) )
		                 return atan2(zn,zd) + pix2;
		              else if ( (cr1<0.0) && (cr2<0.0) && (zn<0.0) )
		                 return atan2(zn,zd) + pix2;
		              else
		                return atan2(zn,zd);
		         }
			 else
		            return 0.0;    /* gammai = atan2(zn,zd) = 0.0 */
	     }
	     else /* x 1----------------2 or 1----------------2 x <==> if Projection of collocation point does not fall in side k */
	       return atan2(zn,zd);
}


double pt_tria_tst(double w, double u, double v, const double *theta, int *theta_id)
{

/* ===purpose=== */

/**********************************************************************************************************************************/
/*************************************************  Point in Triangle Test  *******************************************************/
/***********************  Given a point with barycentric coordinates (u,v) on a plane containing a triangle  **********************/
/************  Determine whether the point (u,v) is outside, or inside, or on an edge, or at a vertex of the triangle  ************/
/**************************  Set the flag theta_id characterizing the position of the query point (u,v)  **************************/
/**********************************************************************************************************************************/

/* Developer: Sylvain Nintcheu Fata, Oak Ridge National Laboratory */
/* Version:   V.1.2                                                */
/* Released:  Apr 14, 2011                                         */


/* Triangle with vertices {y1,y2,y3}

*           y3
*	    *
*	   / \
*	  /   \
*	 /     \
*       /       \
*      /	 \
*     /		  \
*    *-------------*
*  y1               y2

* Oriented edge L1 = [y1,y2] <== Edge1
* Oriented edge L2 = [y2,y3] <== Edge2
* Oriented edge L3 = [y3,y1] <== Edge3
*/


/* Note: (u,v) == Barycentric coordinates of the query point on the plane containing the triangle
 * Note: w = 1-(u+v)
 * Note: The length of vector theta is 3, i.e. theta[3] */
   extern double pi,pix2; /* pi = 3.1415926535897932, pix2 = 2*pi */



     *theta_id = -1;

     if ( u<0.0 || v<0.0 || w<0.0 )
     {
	 *theta_id = 7; return 0.0; /* Query point is Outside the triangle */
     }
     else if ( u==0.0 )
     {
           if ( v==0.0 )
	   {
              *theta_id = 4; return theta[0]; /* Query point is at Vertex1 */
	   }
	   else if ( v==1.0 )
	   {
              *theta_id = 6; return theta[2]; /* Query point is at Vertex3 */
	   }
	   else if ( v<0.0 || v>1.0)
	   {
	      *theta_id = 7; return 0.0; /* Query point is Outside the triangle */
	   }
	   else /* v>0.0 && v<1.0 */
	   {
	      *theta_id = 3; return  pi; /* Query point is on Edge3 of the triangle */
	   }   
     }
     else if ( v==0.0 )
     {
	   if ( u==1.0 )
	   {
              *theta_id = 5; return theta[1]; /* Query point is at Vertex2 */
	   }
	   else if ( u<0.0 || u>1.0 )
	   {
	      *theta_id = 7; return 0.0; /* Query point is Outside the triangle */
	   }
	   else /* u>0.0 && u<1.0 */
	   {
	      *theta_id = 1; return pi; /* Query point is on Edge1 of the triangle */
	   }
     }
     else if ( w==0.0 )
     {
	   if ( u<0.0 || v<0.0 )
	   {
	      *theta_id = 7; return 0.0; /* Query point is Outside the triangle */
	   }
	   else /* u>0.0 && v>0.0 */
	   {
	      *theta_id = 2; return pi; /* Query point is on Edge2 of the triangle */
	   }
     }
     else /* u>0.0 && v>0.0 && w>0.0 */
     {
	*theta_id = 0; return pix2; /* Query point is Inside the triangle */
     }

}


inline double asym_chi(double p1, double p2, double d)
{

/* ===purpose=== */

/**********************************************************************************************************************************/
/************************************  return ASYMTOTIC VALUE of chi when q^2 + et^2 = d << 1  ************************************/
/**********************************************************************************************************************************/

/* Developer: Sylvain Nintcheu Fata, Oak Ridge National Laboratory */
/* Version:   V.1.1                                                */
/* Released:  Mar 21, 2011                                         */


 double z1,z2;
 
 

     if ( fabs(p1)>0.0 && fabs(p2)>0.0 )
     {
           z1 = (d/p1)/p1; z2 = (d/p2)/p2;
           return log( fabs(p2/p1)*( (1.0-0.25*z1*(1.0-0.5*z1))/(1.0-0.25*z2*(1.0-0.5*z2)) ) );
     }
     else return 0.0;

}


void ortho_comp_basis(const double *t1, const double *t2, double *e1, double *e2, double *e3)
{

/* ===purpose=== */

/**********************************************************************************************************************************/
/************************************  Given two linearly independent vectors t1 and t2 in 3D  ************************************/
/** Compute the orthonormal triad {e1,e2,e3} such that e1=t1/norm(t1), norm(e2) = 1 and e2 is orthogonal to e1, and e3 = e1 x e2 **/
/**********************************************************************************************************************************/

/* Developer: Sylvain Nintcheu Fata, Oak Ridge National Laboratory */
/* Version:   V.1.2                                                */
/* Released:  Apr 14, 2011                                         */


/* Note: The length of vectors t1, t2, e1, e2, e3 is 3 */
   int i;
   double al,snrm_te,nrm_te,nrm_tx;



/*********  Orthonormal local basis: e1[3],e2[3],e3[3]  ***************************************************************************/
    snrm_te = t1[0]*t1[0] + t1[1]*t1[1] + t1[2]*t1[2];
    nrm_te = sqrt(snrm_te);

    al = (t1[0]*t2[0] + t1[1]*t2[1] + t1[2]*t2[2])/snrm_te;
    
    for (i=0; i<3; i++) /* Loop over all components */
       e2[i] = t2[i] - al*t1[i];

    nrm_tx = sqrt(e2[0]*e2[0] + e2[1]*e2[1] + e2[2]*e2[2]);

    for (i=0; i<3; i++) /* Loop over all components */
    {   
       e1[i] = t1[i]/nrm_te; e2[i] = e2[i]/nrm_tx;
    }   

/*** Unit Normal vector to the plane spanned by {e1,e2}: e3 = e1 x e2 */
    e3[0] = e1[1]*e2[2] - e2[1]*e1[2];
    e3[1] = e1[2]*e2[0] - e2[2]*e1[0];
    e3[2] = e1[0]*e2[1] - e2[0]*e1[1];

}
